/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mappings are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2004 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */
package jorm;
// Import clauses required by the persistent class jorm.JE2_Reservation
import jorm.JE2_ReservationAccessor;

// Import clauses required by jorm
import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PAccessorGen;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PBindingCtrl;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameGetter;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Date;

public 
	class JE2_ReservationBinding
	extends org.objectweb.jonas_ejb.container.JEntitySwitchCS
	implements PBindingCtrl
    {
    
	/**
	 * The current state of this PBinding object.
	 */
	transient protected byte status;
	/**
	 * The PClassMapping linked to this PBinding object.
	 */
	transient public PClassMapping pClassMapping;
	/**
	 * The PName of this PBinding object.
	 */
	transient protected PName pname;

	public JE2_ReservationBinding() {
	}

	public JE2_ReservationBinding(PClassMapping pcm) {
		pClassMapping = pcm;
	}


	// IMPLEMENTATION OF METHODS FROM THE PBinding INTERFACE

	public void init(PClassMapping pcm) throws PException {
		pClassMapping = pcm;
		status = PBinding.LIFECYCLE_NOTBOUND;
	    pname = null;
	}

	public void write(Object conn, PAccessor pa) throws PException {
		if (pClassMapping == null) {
			throw new PExceptionProtocol("PBinding is not initialized");
		}
		pClassMapping.write(this, conn, pa);
	}

	public void read(Object conn, PAccessor pa) throws PException {
		if (pClassMapping == null) {
			throw new PExceptionProtocol("PBinding is not initialized");
		}
		pClassMapping.read(this, conn, pa);
	}

	public void read(Object conn, PAccessor pa, Object ctx) throws PException {
		if (pClassMapping == null) {
			throw new PExceptionProtocol("PBinding is not initialized");
		}
		pClassMapping.read(this, conn, pa, ctx);
	}

	public boolean exist(Object conn) throws PException {
		if (pClassMapping == null) {
			throw new PExceptionProtocol("PBinding is not initialized");
		}
		return pClassMapping.exist(this, conn);
	}

	public PName export(Object conn) throws PException {
		return pClassMapping.getPBinder().export(conn, this);
	}

	public PName export(Object conn, Object hints) throws PException {
		if (pClassMapping == null) {
			throw new PExceptionProtocol("PBinding is not initialized");
		}
		return pClassMapping.getPBinder().export(conn, this, hints);
	}

	public void unexport(Object conn) throws PException {
		if (pClassMapping == null) {
			throw new PExceptionProtocol("PBinding is not initialized");
		}
        pClassMapping.getPBinder().unexport(conn, this.pname, this);
	}

	public void bind(PName pn) throws PException {
		if (pClassMapping == null) {
			throw new PExceptionProtocol("PBinding is not initialized");
		}
		if (pname!=null) {
			// Unbind the bound Pbinding before
			pClassMapping.getPBinder().unbind(this);
		}
		pClassMapping.getPBinder().bind(pn, this);
	}

	public void unbind() throws PException {
		if (pClassMapping == null) {
			throw new PExceptionProtocol("PBinding is not initialized");
		}
		pClassMapping.getPBinder().unbind(this);
	}

	public PClassMapping getPClassMapping() {
		return pClassMapping;
	}

	public PName getPName() {
		return pname;
	}

	public byte getStatus() {
		return status;
	}

	// IMPLEMENTATION OF METHODS FROM THE PBindingCtrl INTERFACE

	public void setPName(PName pn) {
		pname = pn;
	}

	public void setStatus(byte s) {
		status = s;
	}
}
