/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mappings are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2004 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */
package jorm.rdb;

// Import clauses required by jorm
import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PAccessorGen;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PBindingCtrl;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PClassMappingCtrl;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionNoDSI;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PMappingCallback;
import org.objectweb.jorm.api.PMappingStructuresManager;
import org.objectweb.jorm.api.PNameIterator;
import org.objectweb.jorm.api.PStateGraph;
import org.objectweb.jorm.genclass.api.FieldDesc;
import org.objectweb.jorm.genclass.api.GenClassProp;
import org.objectweb.jorm.lib.EmptyPNameIterator;
import org.objectweb.jorm.lib.MultiplePNameIterator;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.naming.api.PNameManager;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.eval.prefetch.api.PrefetchCache;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.LoggerFactory;
//import for the rdb mapper
import org.objectweb.jorm.api.PPrefetchTupleCollection;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapterException;
import org.objectweb.jorm.mapper.rdb.adapter.api.JoinedTable;
import org.objectweb.jorm.mapper.rdb.lib.RdbExtentGenInfos;
import org.objectweb.jorm.mapper.rdb.lib.RdbTupleCollection;
import org.objectweb.jorm.mapper.rdb.lib.PClassMappingDecoder;
import org.objectweb.jorm.mapper.rdb.lib.PMapperRdb;
import org.objectweb.jorm.mapper.rdb.lib.RdbPMappingStructuresManager;
import org.objectweb.jorm.mapper.rdb.lib.RdbConnectionWrapper;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbMapping;
import org.objectweb.jorm.mapper.rdb.lib.RdbPPolymorphicClass;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.eval.lib.MedorEvaluator;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.eval.api.ConnectionResources;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.CalculatedParameterOperand;
import org.objectweb.medor.expression.lib.BasicParameterOperand;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.query.rdb.lib.BasicRdbStringQueryLeaf;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.datasource.lib.ConnectionFactoryDataStore;
import org.objectweb.medor.expression.lib.BasicBinaryArithmeticOperator;
import org.objectweb.medor.expression.lib.Plus;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.eval.prefetch.api.IntermediaryPrefetchBuffer;
import org.objectweb.medor.eval.prefetch.lib.IntermediaryPrefetchBufferImpl;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.jorm.lib.EncodePName;
import org.objectweb.medor.filter.jorm.lib.CompositePName;
import org.objectweb.medor.filter.jorm.lib.SinglePName;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.optim.lib.IndexesGenerator;
import org.objectweb.perseus.persistence.api.ConnectionHolder;
import org.objectweb.perseus.persistence.api.WorkingSet;
import java.io.IOException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Collections;
import java.util.Collection;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.mapper.rdb.genclass.RdbFieldDesc;
import org.objectweb.jorm.mapper.rdb.genclass.RdbGenClassProp;
import org.objectweb.jorm.mapper.rdb.lib.RdbPrefetchablePCM;


// Import clauses required by the object Olstore_type
import jorm.Olstore_typeAccessor;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

public class Olstore_typeMapping
	extends org.objectweb.jonas_ejb.container.jorm.RdbFactory
	implements PClassMapping, PClassMappingCtrl, Loggable
		, PClassMappingDecoder
		, RdbPrefetchablePCM
    {
	protected static final String MAPPER_NAME = "rdb";
	protected static final String PROJECT_NAME = "jonas";
	//Fields
	protected static final String jonasCMRItem_Type_NAME = "jonasCMRItem_Type";
	protected static final String name_NAME = "name";
	protected static final String properties_NAME = "properties";
	//Hidden fields

	private final static String[][] SUPER_TYPES =  {};

	/** Fully quailied name of the persitent class */
	protected final static String JORM_CLASS_NAME =
		"jorm.Olstore_type";
	/** The actual binder which does the binding action */
	private PBinder binder;
	private PNameCoder classPNC;

	/** The meta object of the current jorm class */
	private Class moc;
	/** The mapper linked to this PClassMapping */
	private PMapper mapper;
	public PNameCoder jonasCMRItem_TypePnc;
	PClassMapping jonasCMRItem_TypeGCM;
	public PNameCoder propertiesPnc;
	PClassMapping propertiesGCM;
	/** the PClassMapping of the sub type */
	private PClassMapping[] subPCMs = null;

	//variables for the resolve method
	private int beginIndex = 0;
	private QueryTree queryTree = null;
	private FieldOperand[] fieldOperands = null;
	private String[] compositeFieldNames = null;
 	/**
	 * The logger instance used by this XMapping instance
	 * and the XBinding instances
	 */
	public Logger logger;
	public LoggerFactory loggerFactory;

	public Olstore_typeMapping() {
 	}
	public PNameCoder getJonasCMRItem_TypePNameCoder() {
	    return jonasCMRItem_TypePnc;
	}
	public PNameCoder getPropertiesPNameCoder() {
	    return propertiesPnc;
	}

	// IMPLEMENTATION OF METHODS FROM THE PClassMappingCtrl INTERFACE

    public void setPNameCoder(PNameCoder pnc)
            throws PException, UnsupportedOperationException {
		throw new UnsupportedOperationException("This class is not a gen class");
    }

    public void setPNameCoder(String fn, PNameCoder pnc)
            throws PException, UnsupportedOperationException {
		if (jonasCMRItem_Type_NAME.equals(fn)) {
			jonasCMRItem_TypePnc = (PNamingContext) pnc;
			return;
		}
		if (properties_NAME.equals(fn)) {
			propertiesPnc = (PNamingContext) pnc;
			return;
		}
		throw new UnsupportedOperationException("This class does'nt contain field reference: " + fn);
    }

    public void setGenClassMapping(PClassMapping pcm)
            throws UnsupportedOperationException {
		throw new UnsupportedOperationException("This class is not a gen class");
    }

    public void setGenClassMapping(String fn, PClassMapping pcm)
            throws UnsupportedOperationException {
		if (jonasCMRItem_Type_NAME.equals(fn)) {
			jonasCMRItem_TypeGCM = pcm;
			return;
		}
		if (properties_NAME.equals(fn)) {
			propertiesGCM = pcm;
			return;
		}
		throw new UnsupportedOperationException("This class does not contain a " + fn + " field referencing a generic class");
    }

    public void setProjectName(String projectname) {
    }


	// IMPLEMENTATION OF METHODS FROM THE PClassMapping INTERFACE

	public void init(PMappingCallback m, MetaObject metaclass) throws PException {
		if (Debug.ON && logger!=null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "");
		mapper = m;
		moc = (Class) metaclass;
		//the getPType method inits the ptype if not already done
		PType ptype = mapper.getPTypeSpace().createPType(
				"jorm.Olstore_type",
				SUPER_TYPES);
		setPType(ptype);
         try {
			adapter = ((PMapperRdb) mapper).getRdbAdapter();
			if (adapter == null) {
				throw new PException("No rdb adapter found on the mapper: " + mapper);
			}

         // Initialization of the gen class /jorm.Olstore_type/properties
        FieldDesc[] fds = null;
		fds = new FieldDesc[1];
			fds[0] = new RdbFieldDesc("jorm.Olstore_type_name", "", PTypeSpace.STRING, "olstore_type_name", adapter.getSqlType(PTypeSpace.STRING.getTypeCode(), true, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE));
		((GenClassProp) propertiesGCM ).setIdentifierDesc(fds);
		fds = new FieldDesc[0];
 		((GenClassProp) propertiesGCM ).setIndexDesc(fds);
		fds = new FieldDesc[2];
			fds[1] = new RdbFieldDesc("jorm.Olstore_properties_JONASAUTOPKFIELD", "", PTypeSpace.OBJINT, "JPK_", adapter.getSqlType(PTypeSpace.OBJINT.getTypeCode(), true, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE));
			fds[0] = new RdbFieldDesc(null, null, PTypeSpace.REFTOP, null, null);
 		((GenClassProp) propertiesGCM).setElemDesc(fds);
			((RdbGenClassProp) propertiesGCM).setTableName("olstore_properties");
			((RdbGenClassProp) propertiesGCM).setColocated(true);
			((RdbGenClassProp) propertiesGCM).setColocatedMaster(false);
			((RdbGenClassProp) propertiesGCM).setReadOnly(true);
			((RdbGenClassProp) propertiesGCM).setTypeConverter(adapter);
        String[] propertiesGCDef = {"java.util.Collection"};
        String[][] propertiesICDef = {};
        ptype = mapper.getPTypeSpace().createPType(
            propertiesGCDef,
            "jorm.Olstore_properties",
            propertiesICDef);
        ((GenClassProp) propertiesGCM).setPType(ptype);
        propertiesPnc.setPType(ptype);
        propertiesGCM.init(m, null);

         // Initialization of the gen class /jorm.Olstore_type/jonasCMRItem_Type
		fds = new FieldDesc[1];
			fds[0] = new RdbFieldDesc("jorm.Olstore_type_name", "", PTypeSpace.STRING, "olstore_type_name", adapter.getSqlType(PTypeSpace.STRING.getTypeCode(), true, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE));
		((GenClassProp) jonasCMRItem_TypeGCM ).setIdentifierDesc(fds);
		fds = new FieldDesc[0];
 		((GenClassProp) jonasCMRItem_TypeGCM ).setIndexDesc(fds);
		fds = new FieldDesc[2];
			fds[1] = new RdbFieldDesc("jorm.Olstore_item_itemId", "", PTypeSpace.OBJINT, "itemId", adapter.getSqlType(PTypeSpace.OBJINT.getTypeCode(), true, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE));
			fds[0] = new RdbFieldDesc(null, null, PTypeSpace.REFTOP, null, null);
 		((GenClassProp) jonasCMRItem_TypeGCM).setElemDesc(fds);
			((RdbGenClassProp) jonasCMRItem_TypeGCM).setTableName("olstore_item");
			((RdbGenClassProp) jonasCMRItem_TypeGCM).setColocated(true);
			((RdbGenClassProp) jonasCMRItem_TypeGCM).setColocatedMaster(false);
			((RdbGenClassProp) jonasCMRItem_TypeGCM).setReadOnly(true);
			((RdbGenClassProp) jonasCMRItem_TypeGCM).setTypeConverter(adapter);
        String[] jonasCMRItem_TypeGCDef = {"java.util.Collection"};
        String[][] jonasCMRItem_TypeICDef = {};
        ptype = mapper.getPTypeSpace().createPType(
            jonasCMRItem_TypeGCDef,
            "jorm.Olstore_item",
            jonasCMRItem_TypeICDef);
        ((GenClassProp) jonasCMRItem_TypeGCM).setPType(ptype);
        jonasCMRItem_TypePnc.setPType(ptype);
        jonasCMRItem_TypeGCM.init(m, null);
 	//register the structures and dependencies
	init(mapper.getPMappingStructuresManager());
        } catch (Exception e) {
			if(e instanceof RdbAdapterException)
				throw new PException(e, "Column type computing error: ");
			else
				e.printStackTrace();
		} 
			computeReadQuery();
		try {
			computeExtentQuery();
		} catch (ExpressionException e) {
			throw new PException(e);
		}
		computePrefetchedExtentQuery();
		}

	public void init(PMappingStructuresManager pmsm) throws PException {
		pmsm.declareClass(JORM_CLASS_NAME);
		try {
			        RdbPMappingStructuresManager msm = (RdbPMappingStructuresManager) pmsm;
        if (mapper == null) {
	       	mapper = msm.getPMapper();
	    }
	    if (adapter == null) {
			adapter = ((PMapperRdb) mapper).getRdbAdapter();
			if (adapter == null) {
				throw new PException("No rdb adapter found on the mapper: " + mapper);
			}
		}
	    // Initialization of the structures of the class jorm.Olstore_type
			        	msm.addTableColumn(JORM_CLASS_NAME, "olstore_type", "name", adapter.getSqlType(PTypeSpace.STRING.getTypeCode(), true, -1, -1), true, true, true);
	    		        // Initialization of the structures of the gen class /jorm.Olstore_type/properties
		        	msm.addTableColumn(JORM_CLASS_NAME, "olstore_properties", "JPK_", adapter.getSqlType(PTypeSpace.OBJINT.getTypeCode(), false, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE), false, false, false);
		        	msm.addTableColumn(JORM_CLASS_NAME, "olstore_properties", "olstore_type_name", adapter.getSqlType(PTypeSpace.STRING.getTypeCode(), false, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE), false, false, false);
			        // Initialization of the structures of the gen class /jorm.Olstore_type/jonasCMRItem_Type
		        	msm.addTableColumn(JORM_CLASS_NAME, "olstore_item", "olstore_type_name", adapter.getSqlType(PTypeSpace.STRING.getTypeCode(), false, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE), false, false, false);
		        	msm.addTableColumn(JORM_CLASS_NAME, "olstore_item", "itemId", adapter.getSqlType(PTypeSpace.OBJINT.getTypeCode(), false, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE), false, false, false);
					//dependencies
	        pmsm.addDependency(JORM_CLASS_NAME, "jorm.Olstore_properties");
	        pmsm.addDependency(JORM_CLASS_NAME, "jorm.Olstore_item");
			} catch(Exception e) {
			throw new PException("Error while registering mapping structures: " + e.getMessage());
		}
		classDefined(pmsm);
	}
	
	protected void classDefined() throws PException {
		//register class in the cluster
        classDefined(mapper.getPMappingStructuresManager());
	}

	public void classDefined(PMappingStructuresManager pmsm) throws PException {
		if (pmsm == null) {
			throw new PException("PMappingStructureManager is null: " + JORM_CLASS_NAME + " cannot be registered in a cluster.");
		}
		//register class in the cluster
        pmsm.classDefined(JORM_CLASS_NAME);
	}
	
	public PMapper getPMapper() {
		return mapper;
	}

	public PBinder getPBinder() {
		return binder;
	}

	public PNameCoder getClassPNameCoder() {
		return classPNC;
	}

	/** This method is used only in the case of the current
	 * class is a gen class and contains references*/
	public PNameCoder getPNameCoder()
		throws UnsupportedOperationException {
		throw new UnsupportedOperationException("This class is not a gen class");
	}

	public PNameCoder getPNameCoder(String fn)
		throws UnsupportedOperationException {
			if (jonasCMRItem_Type_NAME.equals(fn))
				return jonasCMRItem_TypePnc;
			if (properties_NAME.equals(fn))
				return propertiesPnc;
		throw new UnsupportedOperationException("This class does'nt contain field reference " + fn);
	}

	public String getClassName() {
		return JORM_CLASS_NAME;
	}

	/** This method is used only in the case of the current
	 * class is a gen class and contains another gen class*/
	public PClassMapping getGenClassMapping() throws UnsupportedOperationException {
		throw new UnsupportedOperationException("This class is not a gen class");
	}

	public PClassMapping getGenClassMapping(String fn)
		throws UnsupportedOperationException {
			if (jonasCMRItem_Type_NAME.equals(fn))
				return jonasCMRItem_TypeGCM;
			if (properties_NAME.equals(fn))
				return propertiesGCM;
			throw new UnsupportedOperationException("This class does not contain a " + fn + " field which would reference a generic class");
	}

	public boolean isConform(String mapperName) {
		return mapperName.startsWith(MAPPER_NAME);
	}

	public String getProjectName() {
		return PROJECT_NAME;
	}

	public MetaObject getMetaInfo() {
		return moc;
	}

	public String[][] getSuperClassNames() {
		return SUPER_TYPES;
	}

	public void setPBinder(PBinder b) {
		binder = b;
	}

	public void setClassPNameCoder(PNameCoder pnc) throws PException {
		classPNC = pnc;
	}

	public PBinding createPBinding() throws PException {
		if (Debug.ON && logger!=null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "");
		PBinding binding = new jorm.Olstore_typeBinding();
		binding.init(this);
		return binding;
	}
	/** @deprecated */
    public Iterator getPNameIterator(Object conn) throws PException {
		return getPNameIterator(conn, false, false, null);
    }

	public PNameIterator getPNameIterator(Object conn,
	                                      boolean withSubType,
	                                      boolean prefetching,
	                                      Object txctx) throws PException {
		PNameIterator it = null;
		Connection connection = RdbConnectionWrapper.narrow2SQL(conn);
		try {
			String mapperName = mapper.getMapperName();
			if(mapperName.indexOf(".") != -1)
				mapperName = mapperName.substring(0, mapperName.indexOf("."));	
			RdbMapping rdbm = (RdbMapping) mapper.getMetaInfoManager().getClass(this.getClassName()).getClassProject(this.getProjectName()).getMapping(mapperName);
			RdbClassMapping rdbcm = (RdbClassMapping) rdbm.getClassMapping();
			String queryToExecute = "";
			int beginIndex = 0;
			if(withSubType){
				//the list of fields belonging to the primary key
				String pk = "olstore_type.name";
				queryToExecute = rdbcm.getRdbInheritanceQuery().getExtentQuery(mapper,prefetching, true, pk, this);	
				//System.out.println("query to execute:" + queryToExecute + " compared with the generated query: " + (prefetching ? prefetchedExtentQuery : extentQuery));
				beginIndex =  rdbcm.getRdbInheritanceQuery().getExtentFieldsNames(prefetching, true).size() + 1;
			}
			else{
				queryToExecute = (prefetching ? prefetchedExtentQuery : extentQuery);
				//System.out.println("query to execute:" + queryToExecute);
				//compute the begining index for the PName: number of fields to select + 1
				beginIndex = 1 + 1;
			}
			PreparedStatement ps = connection.prepareStatement(queryToExecute);
			it = new Olstore_typeResultIterator(txctx, ps, ps.executeQuery(), binder, prefetching, withSubType, logger, beginIndex, this);
        } catch (SQLException e) {
            throw new PExceptionIO(e, "SQL problem while computing the PName Iterator: class <Olstore_type>.");
        } catch (MedorException e) {
            throw new PExceptionIO(e, "Medor problem while initializing the PName Iterator: class <Olstore_type>.");
        }
				return it;
			}

	public PClassMapping[] getSubPCMs() throws PException {
		PType ptype = getPType();
		if (ptype == null) {
			throw new PException("No PType defined for the PClassMapping of the class" + JORM_CLASS_NAME);
		}
		PType[] subTypes = ptype.getSubTypes();
		if (subTypes == null || subTypes.length == 0) {
			//No direct subType are mapped then return null
			return null;
		}
		if (subPCMs == null || subPCMs.length != subTypes.length) {
			//compute the PCM array
			PMapper mapper = getPMapper();
			subPCMs = new PClassMapping[subTypes.length];
			for (int i=0; i<subTypes.length; i++) {
				subPCMs[i] = mapper.lookup(subTypes[i].getJormName());
			}
		}
		return subPCMs;
	}
	
	// ADD YO //
	public synchronized PName resolve(Object conn, PName pname) throws PException {
				//mapper rdb but not polymorphic
		return pname;
	 	}
	
	public boolean match(Object obj, boolean intermediaryTuple) throws PException {
					return false;
				}
	
	public PName getDecodedPName(TupleCollection tc, PName pname, boolean intermediaryTuple) throws PException {
		PBinder binder = null;
	//test if the tuple is an instance of this class
	if (this.match(tc, intermediaryTuple)) {
		binder  = this.getPBinder();
	} else {
		//test for all the children
		boolean match = false;
		for (int i = 0; i<subPCMs.length && !match; i++) {
				match = subPCMs[i].match(tc, intermediaryTuple);
				binder = subPCMs[i].getPBinder();
		}
		//if no pcm found, error
		if (!match)
			throw new PException("Problem with Medor while resolving pname on the PCM " + JORM_CLASS_NAME + ". No class found for the pname " + pname.toString() + ".");
	}
			return binder.decodeString(pname.encodeString());
		}
	// END ADD YO //
	
    public void configureRefFields(ReferenceConfigurator rc) throws PException, UnsupportedOperationException {
        PClassMapping gcm = null;
        jonasCMRItem_TypePnc = (PNamingContext) rc.getPNameCoder(JORM_CLASS_NAME, jonasCMRItem_Type_NAME, new String[]{"java.util.Collection"});
        jonasCMRItem_TypeGCM = rc.getGenClassMapping(JORM_CLASS_NAME, jonasCMRItem_Type_NAME, new String[]{"java.util.Collection"}, "jorm.Olstore_item");
        gcm = jonasCMRItem_TypeGCM;
        propertiesPnc = (PNamingContext) rc.getPNameCoder(JORM_CLASS_NAME, properties_NAME, new String[]{"java.util.Collection"});
        propertiesGCM = rc.getGenClassMapping(JORM_CLASS_NAME, properties_NAME, new String[]{"java.util.Collection"}, "jorm.Olstore_properties");
        gcm = propertiesGCM;
    }

	public PType getPType() {
		return  binder.getPType();
	}

	public void setPType(PType pt) {
		binder.setPType(pt);
	}


	public void write(PBinding pb, Object conn, PAccessor pa)
		throws PException {
		Olstore_typeAccessor paccessor = null;
		PAccessorGen paccessorGen = null;
		if (pa == null)
			throw new PExceptionProtocol("Accessor must not be null");
		else if (pa instanceof Olstore_typeAccessor)
			paccessor = (Olstore_typeAccessor) pa;
		else if (pa instanceof PAccessorGen)
			paccessorGen = (PAccessorGen) pa;
		else
			throw new PExceptionProtocol("Wrong type of accessor: " + pa);
		byte status = pb.getStatus();
		byte nextState =
			PStateGraph.nextStatePBinding(status, PBinding.ACTION_WRITE);
		if (nextState == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized WRITE operation, status=" + status);
		Connection connection = RdbConnectionWrapper.narrow2SQL(conn);
		try {
			String sqlquery;
		    if (status == PBinding.LIFECYCLE_DELTOWRITE) {
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "Remove");
//Table olstore_type : readOnly=false mustGenerateUpdate=false colocatedTable=false joins=$tab.joins
                sqlquery = deleteolstore_typeTable(connection, conn, paccessorGen, paccessor, pb.getPName());
                if (sqlquery != null) {
                    throw new PExceptionIO("Risk to loose modifications to class [Olstore_type]: problematic SQL query -> {" + sqlquery + "}");
                }
            } else if (status==PBinding.LIFECYCLE_NEWTOWRITE) {
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "Insert");
                sqlquery = insertolstore_typeTable(connection, conn, paccessorGen, paccessor, pb.getPName());
                if (sqlquery != null) {
                    throw new PExceptionIO("Risk to loose modifications to class [Olstore_type]: problematic SQL query -> {" + sqlquery + "}");
                }
		    } else if (status==PBinding.LIFECYCLE_ACTIVEFORIO) {
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "Update");
		    }
		} catch (java.sql.SQLException se) {
			throw new PExceptionIO(se, "SQL problem while writing the DSI.");
		} catch (java.io.IOException ioe) {
			throw new PExceptionIO(ioe, "Stream IO problem while writing the DSI.");
		} catch (RdbAdapterException rae) {
			throw new PExceptionIO(rae, "SQL problem while writing the DSI.");
		}
		// Changing the current state
		((PBindingCtrl) pb).setStatus(nextState);
	}

	public void read(PBinding pb, Object conn, PAccessor pa) throws PException {
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "Read");
		byte status = pb.getStatus();
		Olstore_typeAccessor paccessor = null;
		PAccessorGen paccessorGen = null;
		if (pa == null)
			throw new PExceptionProtocol("Accessor must not be null");
		else if (pa instanceof Olstore_typeAccessor)
			paccessor = (Olstore_typeAccessor) pa;
		else if (pa instanceof PAccessorGen)
			paccessorGen = (PAccessorGen) pa;
		else
			throw new PExceptionProtocol("Wrong type of accessor: " + pa);
		byte nextState =
			PStateGraph.nextStatePBinding(status, PBinding.ACTION_READ);
		if (nextState==PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized READ operation, status=" + status);
		PName pname = pb.getPName();
		Connection connection = RdbConnectionWrapper.narrow2SQL(conn);
        PreparedStatement pstmt = null;
        ResultSet resultSet = null;
        PException resex = null;
		try {
            pstmt = connection.prepareStatement(readQuery);
            adapter.setString(pstmt, 1, pname.encodeString());
            resultSet = pstmt.executeQuery();
            if(resultSet.next() == false)
                throw new PExceptionNoDSI(
                    "No entry matching with the following identifier: " + pname);
            PName refname;
            Object fieldval;
            long notnullval;
			if (paccessor != null) {
				//read public primitive fields
                paccessor.paSetName(adapter.getString(resultSet, 1, null));
				//decode PName for references
				//writeReadRef  isSpecific=true gen_ref=org.objectweb.jorm.mapper.rdb.generator.RdbGenRef@e0e515 / cnId=-1 / cnPackage=null / cnFieldColumns=null / refColumn=org.objectweb.jorm.mapper.rdb.generator.RdbGenColumn@dfcb47 / nd=org.objectweb.jorm.metainfo.lib.BasicNameDef@180cf2a / fieldName=jonasCMRItem_Type / join=null
				//fieldName=name
                fieldval = adapter.getString(resultSet, 1, null);
                if (resultSet.wasNull()) {
                    refname = jonasCMRItem_TypePnc.decodeString(null);
                } else {
                    refname = jonasCMRItem_TypePnc.decodeString((java.lang.String) fieldval);
                }
                paccessor.paSetJonasCMRItem_Type(refname, conn);
				//writeReadRef  isSpecific=true gen_ref=org.objectweb.jorm.mapper.rdb.generator.RdbGenRef@1c1c92b / cnId=-1 / cnPackage=null / cnFieldColumns=null / refColumn=org.objectweb.jorm.mapper.rdb.generator.RdbGenColumn@dfcb47 / nd=org.objectweb.jorm.metainfo.lib.BasicNameDef@1a0d866 / fieldName=properties / join=null
				//fieldName=name
                fieldval = adapter.getString(resultSet, 1, null);
                if (resultSet.wasNull()) {
                    refname = propertiesPnc.decodeString(null);
                } else {
                    refname = propertiesPnc.decodeString((java.lang.String) fieldval);
                }
                paccessor.paSetProperties(refname, conn);
			} else if (paccessorGen != null) {
				//read public primitive fields
                paccessorGen.paSetStringField(name_NAME, adapter.getString(resultSet, 1, null));
				//decode PName for references
				//writeReadRef  isSpecific=false gen_ref=org.objectweb.jorm.mapper.rdb.generator.RdbGenRef@1c1c92b / cnId=-1 / cnPackage=null / cnFieldColumns=null / refColumn=org.objectweb.jorm.mapper.rdb.generator.RdbGenColumn@dfcb47 / nd=org.objectweb.jorm.metainfo.lib.BasicNameDef@1a0d866 / fieldName=properties / join=null
				//fieldName=name
                fieldval = adapter.getString(resultSet, 1, null);
                if (resultSet.wasNull()) {
                    refname = propertiesPnc.decodeString(null);
                } else {
                    refname = propertiesPnc.decodeString((java.lang.String) fieldval);
                }
                paccessorGen.paSetRefField(properties_NAME, refname, conn);
				//writeReadRef  isSpecific=false gen_ref=org.objectweb.jorm.mapper.rdb.generator.RdbGenRef@1c1c92b / cnId=-1 / cnPackage=null / cnFieldColumns=null / refColumn=org.objectweb.jorm.mapper.rdb.generator.RdbGenColumn@dfcb47 / nd=org.objectweb.jorm.metainfo.lib.BasicNameDef@1a0d866 / fieldName=properties / join=null
				//fieldName=name
                fieldval = adapter.getString(resultSet, 1, null);
                if (resultSet.wasNull()) {
                    refname = propertiesPnc.decodeString(null);
                } else {
                    refname = propertiesPnc.decodeString((java.lang.String) fieldval);
                }
                paccessorGen.paSetRefField(properties_NAME, refname, conn);
			}
		} catch (PException pe) {
			resex = pe;
		} catch (Exception pe) {
			resex = new PExceptionIO(pe, "Problem while reading the DSI.");
		} finally {
		    try {
		        if (resultSet != null) {
			        resultSet.close();
			        resultSet = null;
			    }
			    if (pstmt != null) {
			        pstmt.close();
			        pstmt = null;
			    }
			} catch (SQLException se) {
				if (resex == null) {
			    	resex = new PExceptionIO(se, "Problem while closing ResultSet or Statment");
			    }
			}
			if (resex != null) {
				throw resex;
			}
		}
		((PBindingCtrl) pb).setStatus(nextState);
	}

	public void read(PBinding pb, Object conn, PAccessor pa, Object txctx) throws PException {
	    PrefetchCache pc = getPMapper().getPrefetchCache();
	    if (pc == null || txctx == null) {
	        read(pb, conn, pa);
	        return;
	    }
		PName pname = pb.getPName();
	    try {
	        Tuple tuple = pc.getPrefetchTuple(pname, this, txctx);
	        if (tuple == null) {
	            read(pb, conn, pa);
	            return;
	        }
		    Olstore_typeAccessor paccessor = null;
		    PAccessorGen paccessorGen = null;
		    if (pa == null)
			    throw new PExceptionProtocol("Accessor must not be null");
		    else if (pa instanceof Olstore_typeAccessor)
			    paccessor = (Olstore_typeAccessor) pa;
		    else if (pa instanceof PAccessorGen)
			    paccessorGen = (PAccessorGen) pa;
		    else
			    throw new PExceptionProtocol("Wrong type of accessor: " + pa);
			byte status = pb.getStatus();
		    byte nextState = PStateGraph.nextStatePBinding(status, PBinding.ACTION_READ);
		    if (nextState==PBinding.LIFECYCLE_ERROR)
			    throw new PExceptionProtocol("Unauthorized READ operation, status=" + status);
            Object fval;
            long notnullval;
            if (paccessor != null) {
				fval =
					tuple.getString(1);
                if (tuple.isDefined(1)) {
                    paccessor.paSetJonasCMRItem_Type(jonasCMRItem_TypePnc.decodeString((java.lang.String) fval), conn);
                } else {
                    paccessor.paSetJonasCMRItem_Type(jonasCMRItem_TypePnc.getNull(), conn);
                }
                paccessor.paSetName(tuple.getString(1));
				fval =
					tuple.getString(1);
                if (tuple.isDefined(1)) {
                    paccessor.paSetProperties(propertiesPnc.decodeString((java.lang.String) fval), conn);
                } else {
                    paccessor.paSetProperties(propertiesPnc.getNull(), conn);
                }
            } else if (paccessorGen != null) {
				fval =
					tuple.getString(1);
                if (tuple.isDefined(1)) {
                    paccessorGen.paSetRefField(jonasCMRItem_Type_NAME, jonasCMRItem_TypePnc.decodeString((java.lang.String) fval), conn);
                } else {
                    paccessorGen.paSetRefField(jonasCMRItem_Type_NAME, jonasCMRItem_TypePnc.getNull(), conn);
                }
                paccessorGen.paSetStringField(name_NAME, tuple.getString(1));
				fval =
					tuple.getString(1);
                if (tuple.isDefined(1)) {
                    paccessorGen.paSetRefField(properties_NAME, propertiesPnc.decodeString((java.lang.String) fval), conn);
                } else {
                    paccessorGen.paSetRefField(properties_NAME, propertiesPnc.getNull(), conn);
                }
            }
		((PBindingCtrl) pb).setStatus(nextState);
		} catch (MedorException me) {
		    throw new PExceptionProtocol(me, "Problem while trying to prefetch data from Medor");
		}
    }

	public boolean exist(PBinding pb, Object conn) throws PException {
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) logger.log(BasicLevel.DEBUG, "Exist");
		byte status = pb.getStatus();
		byte nextState =
			PStateGraph.nextStatePBinding(status, PBinding.ACTION_EXIST);
		if (nextState==PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized EXIST operation, status=" + status);
		((PBindingCtrl) pb).setStatus(nextState);
		PName pname = pb.getPName();
		Connection connection = RdbConnectionWrapper.narrow2SQL(conn);
        PreparedStatement pstmt = null;
        ResultSet rs = null;
        PException resex = null;
        boolean exist = false;
		try {
            pstmt = connection.prepareStatement("SELECT olstore_type.name FROM olstore_type WHERE olstore_type.name = ?");
            adapter.setString(pstmt, 1, pname.encodeString());
			rs = pstmt.executeQuery();
            exist = rs.next();
		} catch (java.sql.SQLException se) {
			resex = new PExceptionIO(se, "SQL problem while testing the existence of a DSI.");
		} finally {
		    try {
			    if (rs != null) {
			        rs.close();
			        rs = null;
			    }
			    if (pstmt != null) {
			        pstmt.close();
			        pstmt = null;
			    }
			} catch (SQLException se) {
				if (resex == null) {
					resex = new PExceptionIO(se, "Problem while closing ResultSet or Statment");
			    }
			}
			if (resex != null) {
				throw resex;
			}
		}
		return exist;
	}

	//the map of [pclassmapping->table of int]
	//build dynamically
	private HashMap associationTable =  new HashMap();
	
	public HashMap getAssociationTable(){
		return associationTable;
	}
	
	public void addAssociation(PClassMapping targetClass, int[] indexes){
		associationTable.put(targetClass, indexes);
	}
	
	public int[] getIndexesTable(PClassMapping targetClass){
		return (int[]) associationTable.get(targetClass);
	}
	
    private String insertolstore_typeTable(Connection connection,
                                                Object conn,
                                                PAccessorGen paccessorGen,
                                                Olstore_typeAccessor paccessor,
    											PName pname)
        throws PException, SQLException, IOException, RdbAdapterException {

		Object nullField = null;
        String res = null;
        PreparedStatement pstmt = null;
        try {
            if (paccessor != null) {
                pstmt = connection.prepareStatement(res = "INSERT INTO olstore_type (name) VALUES (?)");
				//name
                nullField = paccessor.paGetName();
                if (nullField == null) {
                    adapter.setNull(pstmt, 1, adapter.getSqlTypeCode(PTypeSpace.STRING.getTypeCode(), null));
                } else {
                   	adapter.setString(pstmt, 1, ((java.lang.String) nullField));
                }
            } else if (paccessorGen != null) {
                pstmt = connection.prepareStatement(res = "INSERT INTO olstore_type (name) VALUES (?)");
				//name
                nullField = paccessorGen.paGetStringField(name_NAME);
                if (nullField == null) {
                    adapter.setNull(pstmt, 1, adapter.getSqlTypeCode(PTypeSpace.STRING.getTypeCode(), null));
                } else {
                   	adapter.setString(pstmt, 1, ((java.lang.String) nullField));
                }
            }
            if (pstmt.executeUpdate() == 1) {
                res = null;
            }
        } finally {
            if (pstmt != null) {
                pstmt.close();
                pstmt = null;
            }
        }
        return res;
    }
    private String deleteolstore_typeTable(
    	Connection connection,
    	Object conn,
    	PAccessorGen paccessorGen,
    	Olstore_typeAccessor paccessor,
    	PName pname) throws PException, SQLException, RdbAdapterException {
		Object nullField = null;
        String res = null;
        PreparedStatement pstmt = null;
        try {
            pstmt = connection.prepareStatement(res = "DELETE FROM olstore_type WHERE name = ?");
				//name
                nullField = paccessor.paGetName();
                if (nullField == null) {
                    adapter.setNull(pstmt, 1, adapter.getSqlTypeCode(PTypeSpace.STRING.getTypeCode(), null));
                } else {
                   	adapter.setString(pstmt, 1, ((java.lang.String) nullField));
                }
            if (pstmt.executeUpdate() == 1) {
                res = null;
            }
        } finally {
            if (pstmt != null) {
                pstmt.close();
                pstmt = null;
            }
        }
        return res;
    }

	/**
	 * Is the RdbAdapter to use for managing the various relational database.
	 * This field must not be null after the mapping of the class
	 */
	public RdbAdapter adapter;
	/**
	 * Is the select query to use in the Binding.read method
	 * The building of the from clause depends on database (outer join syntax).
	 * The RdbAdapter is use to built it. The read query is built once time
	 * when the class is mapped.
	 */
	public String readQuery = null;

	/**
	 * Is the select query to use in the Mapping.getPNameIterator method
	 * in the case where the prefetching is activated.
	 * The building of the from clause depends on database (outer join syntax).
	 * The RdbAdapter is use to built it. The extent query is built once time
	 * when the class is mapped.
	 */
	public String prefetchedExtentQuery = null;

	/**
	 * Is the select query to use in the Mapping.getPNameIterator method
	 * in the case where the prefetching is NOT activated.
	 * The building of the from clause depends on database (outer join syntax).
	 * The RdbAdapter is use to built it. The extent query is built once time
	 * when the class is mapped.
	 */
	public String extentQuery = null;

	/**
	 * Building of the structure defining the FROM clause with
	 *  JoinedTable instances
	 */
	private JoinedTable getJoinedTable() {
		JoinedTable jt = new JoinedTable("olstore_type");
		JoinedTable.Join jtj = null;
        return jt;
	}

    public String[] getPrefetchSelectCols() {
		return new String[]{"olstore_type.name"};
	}

	public String getPrefetchWhereClause() {
		return null;
	}

	public java.util.Properties getCNId2Column() {
		java.util.Properties p = new java.util.Properties();
		p.setProperty("", "olstore_type.name");
		return p;
	}

	public List getPrefetchTables() {
		return Collections.singletonList(getJoinedTable());
	}

	/**
	 * Compute the read query with the help of the RdbAdapter
	 */
	public void computeReadQuery() {
		readQuery = adapter.getQuery("olstore_type.name",
			Collections.singletonList(getJoinedTable()),
			"olstore_type.name = ?",
			false, false);
	}

	/**
	 * Compute the extent query (without prefetch) with the help of the
	 * RdbAdapter
	 */
	public void computeExtentQuery() throws ExpressionException {
		if (extentQuery == null) {
			extentQuery = adapter.getQuery("olstore_type.name",
				Collections.singletonList(getJoinedTable()),
			null,
			false, false);
		}
	}

	/**
	 * Compute the extent query (with prefetch) with the help of the RdbAdapter
	 */
	public void computePrefetchedExtentQuery() {
		if (prefetchedExtentQuery == null) {
			prefetchedExtentQuery = adapter.getQuery("olstore_type.name, olstore_type.name" + adapter.getColumnAliasExpr("pk0") + "",
			Collections.singletonList(getJoinedTable()),
			null,
			false, false);
		}

	}


	public PName decode(ResultSet rs, int startindex, Object png) throws IOException, SQLException, PException {
        return classPNC.decodeString(adapter.getString(rs, startindex, null));
	}



	
	// IMPLEMENTATION OF METHODS FROM THE Loggable INTERFACE

	public void setLogger(Logger l) {
		logger = l;
		((Loggable) jonasCMRItem_TypeGCM).setLogger(l);
		((Loggable) propertiesGCM).setLogger(l);
	}

	public void setLoggerFactory(LoggerFactory lf) {
		loggerFactory = lf;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		return loggerFactory;
	}
}




class Olstore_typeResultIterator implements PNameIterator {
    ResultSet rs;
	private PBinder binder;
	private Olstore_typeMapping pclassMapping;
    private boolean hasNext;
    private boolean isForw;
	private PreparedStatement ps;
	private Logger logger;

    private boolean prefetching;
    
    private PPrefetchTupleCollection tc = null;
    private int pnameIndex;

	Olstore_typeResultIterator(Object txctx,
		PreparedStatement ps,
		ResultSet r,
		PBinder binder,
		boolean prefetching,
		boolean withSubType,
		Logger l,
		int pi,
		Olstore_typeMapping pcm) throws SQLException, MedorException {

		this.ps = ps;
		this.binder = binder;
		this.prefetching = prefetching;
        rs = r;
        logger = l;
        pclassMapping = pcm;
        if (!prefetching) {
        	pnameIndex = 1;
        } else {
        	pnameIndex = pi;
        	tc = new RdbTupleCollection(txctx, r, ps, pnameIndex, logger, pclassMapping, withSubType, this);
        }
        hasNext = (!prefetching) ? rs.next() : tc.next();
        isForw = true;
    }

	public void finalize() throws Throwable {
		if (rs != null)
			rs.close();
		if (ps != null)
			ps.close();
		super.finalize();
	}

    public void close() {
        if (!prefetching) {
            if (rs!= null) {
                try {
                    rs.close();
                } catch(Exception e) {
                }
                rs = null;
            }
            hasNext = false;
            isForw = false;
        }
    }

    public boolean hasNext() {
    	boolean res;
        if (isForw) {
            res = hasNext;
        } else {
            isForw = true;
            try {
                hasNext = (!prefetching) ? rs.next() : tc.next();
                res = hasNext;
            } catch(Exception e) {
                e.printStackTrace();
                res = false;
            }
        }
        if (res == false) {
        	try {
				if (rs != null) {
					rs.close();
					rs = null;
				}
				if (ps != null) {
					ps.close();
					ps = null;
				}
			} catch (SQLException e) {
				rs = null;
				ps = null;
			}
        }
        return res;
    }

    public void remove() throws UnsupportedOperationException, IllegalStateException {
        throw new UnsupportedOperationException();
    }

    public Object next() throws NoSuchElementException {
        try {
            if (!isForw) {
                hasNext = (!prefetching) ? rs.next() : tc.next();
                isForw = true;
            }
            if (!hasNext) {
                throw new NoSuchElementException();
            }
            isForw = false;
            if (prefetching) {
            	tc.getPrefetchBuffer().addPrefetchTuple();
            }
            return pclassMapping.decode(rs, pnameIndex, null);
        } catch (NoSuchElementException e) {
        	throw e;
        } catch (Exception e) {
        	if (logger != null) {
        		logger.log(BasicLevel.ERROR, "Impossible to fetch the next element", e);
        	}
            throw new NoSuchElementException(
                    "Impossible to decode to an ID PName: " + e.getMessage());
        }
    }
}












